@@ -47,6 +47,18 @@ void uwsgi_register_hook(char *name, int (*func)(char *)) {
 	}
 }
 
+static int uwsgi_hook_alarm(char *arg) {
+	char *space = strchr(arg,' ');
+	if (!space) {
+		uwsgi_log("invalid alarm hook syntax, must be: <alarm> <msg>\n");
+		return -1;
+	}
+	*space = 0;
+	uwsgi_alarm_trigger(arg, space+1,  strlen(space+1));
+	*space = ' ';
+	return 0;
+}
+
 static int uwsgi_hook_chdir(char *arg) {
 	int ret = chdir(arg);
 	if (ret) {
@@ -154,6 +166,17 @@ static int uwsgi_hook_write(char *arg) {
         return 0;
 }
 
+static int uwsgi_hook_creat(char *arg) {
+        int fd = open(arg, O_WRONLY|O_CREAT|O_TRUNC, 0666);
+        if (fd < 0) {
+                uwsgi_error_open(arg);
+                return -1;
+        }
+        close(fd);
+        return 0;
+}
+
+
 static int uwsgi_hook_append(char *arg) {
         char *space = strchr(arg, ' ');
         if (!space) {
@@ -262,6 +285,20 @@ static int uwsgi_hook_chmod(char *arg) {
 	return ret;
 }
 
+static int uwsgi_hook_sticky(char *arg) {
+	struct stat st;
+	if (stat(arg, &st)) {
+                uwsgi_error("uwsgi_hook_sticky()/stat()");
+		return -1;
+	}
+        if (chmod(arg, st.st_mode | S_ISVTX)) {
+                uwsgi_error("uwsgi_hook_sticky()/chmod()");
+		return -1;
+        }
+        return 0;
+}
+
+
 static int uwsgi_hook_chown(char *arg) {
         char *space = strchr(arg, ' ');
         if (!space) {
@@ -342,6 +379,9 @@ static int uwsgi_hook_chown2(char *arg) {
 }
 
 
+#ifdef __sun__
+extern int sethostname(char *, int);
+#endif
 static int uwsgi_hook_hostname(char *arg) {
 #ifdef __CYGWIN__
 	return -1;
@@ -456,8 +496,13 @@ void uwsgi_register_base_hooks() {
 	uwsgi_register_hook("chown", uwsgi_hook_chown);
 	uwsgi_register_hook("chown2", uwsgi_hook_chown2);
 
+	uwsgi_register_hook("sticky", uwsgi_hook_sticky);
+
 	uwsgi_register_hook("exec", uwsgi_hook_exec);
 
+	uwsgi_register_hook("create", uwsgi_hook_creat);
+	uwsgi_register_hook("creat", uwsgi_hook_creat);
+
 	uwsgi_register_hook("write", uwsgi_hook_write);
 	uwsgi_register_hook("writen", uwsgi_hook_writen);
 	uwsgi_register_hook("append", uwsgi_hook_append);
@@ -476,9 +521,12 @@ void uwsgi_register_base_hooks() {
 
 	uwsgi_register_hook("hostname", uwsgi_hook_hostname);
 
+	uwsgi_register_hook("alarm", uwsgi_hook_alarm);
+
 	// for testing
 	uwsgi_register_hook("exit", uwsgi_hook_exit);
 	uwsgi_register_hook("print", uwsgi_hook_print);
+	uwsgi_register_hook("log", uwsgi_hook_print);
 }
 
 void uwsgi_hooks_run(struct uwsgi_string_list *l, char *phase, int fatal) {
